"""Hardware control for Manson PSU."""
import serial


class MansonPSU:
    """Control Manson Power Supply Unit.

    :param port: Serial port (e.g., "COM3" or "/dev/ttyUSB0")
    :param baudrate: Communication speed, typically 9600 for Manson PSU
    :param timeout: Timeout for serial communication in seconds

    """
    def __init__(self, port: str, baudrate: int = 9600, timeout: int = 1):
        """Initialize the serial connection to the Manson power supply unit."""
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self._serial = self.init_connection()

    def init_connection(self):
        """Open the serial connection."""
        return serial.Serial(
            port=self.port,
            baudrate=self.baudrate,
            timeout=self.timeout,
            bytesize=serial.EIGHTBITS,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE
        )

    def disconnect(self):
        """Close the serial connection."""
        if self._serial and self._serial.is_open:
            self._serial.close()

    def send_command(self, command: str) -> str:
        """Send a command to the PSU and return the response.

        :param command: command to send
        :return: response from the PSU
        """
        command = command.strip() + '\r'
        with self._serial as connection:
            connection.write(command.encode('ascii'))
            response = connection.readline().decode('ascii').strip()

        return response

    def set_voltage(self, voltage: float) -> bool:
        """Set the output voltage.

        :param voltage: desired voltage
        :return: PSU response as boolean
        """
        if voltage <= 0 or voltage >=100:
            raise ValueError

        command = f"VOLT{int(voltage * 10):03}"
        response = self.send_command(command)
        return response != ""

    def set_current(self, current: float) -> bool:
        """Set the output current.

        :param current: desired current
        :return: PSU response as boolean
        """
        if current < 0 or current >= 10:
            raise ValueError

        command = f"CURR{int(current * 100):03}"
        response = self.send_command(command)
        return response != ""

    def get_voltage(self) -> float:
        """Get the current output voltage.

        :return: output voltage
        """
        response = self.send_command("GETD")
        if response != '':
            return int(response[0:4]) / float(100)
        return False

    def get_current(self) -> float:
        """Get the current output current.

        :return: output current
        """
        response = self.send_command("GETD")
        if response != '':
            return int(response[4:8]) / float(100)
        return False

    def set_output(self, state: bool) -> bool:
        """Turn the output on or off.

        :param state: True to turn on, False to turn off
        :return: PSU response as boolean
        """
        on, off = "SOUT0", "SOUT1"
        command = on if state else off
        response = self.send_command(command)
        return response != ""



"""Hardware control module - dedicated to controlling real hardware through a unified API."""
import time
from typing import List

from ._base import Signal, Signals, SignalProperties, DISignal, DOSignal, SignalsEngine


def signal_power_cycle(relays_board: Signals, do_signal_names: List[str], cutoff_time_s=0.25):
    """Switch off, wait, switch on DO signal by given name.

    :param relays_board: signal instance
    :param do_signal_names: names of the signal on which the power cycle will be performed
    :param cutoff_time_s: time how long the signal will be off

    """
    relays_board.read_states()
    for do_signal_name in do_signal_names:
        signal = getattr(relays_board, do_signal_name)
        signal.state = False
    relays_board.write_states()
    time.sleep(cutoff_time_s)
    for do_signal_name in do_signal_names:
        signal = getattr(relays_board, do_signal_name)
        signal.state = True
    relays_board.write_states()



"""Provide signals support for `WaveShare Modbus Poe ETH Relay` module.

Example configuration:

.. code-block:: python

    import time

    from hwc.common import Signals, DOSignal
    from hwc.waveshare import (
        SignalEnginWaveShareEthMb, SignalPropertiesWaveShareEthMb)


    class PowerBoard(Signals):
        dut_1_power = DOSignal(
            hardware_properties=[
                SignalPropertiesWaveShareEthMb(relay_no=1, active_state=True),
            ]
        )
        dut_2_power = DOSignal(
            hardware_properties=[
                SignalPropertiesWaveShareEthMb(relay_no=2, active_state=False),
            ]
        )


    engine = SignalEnginWaveShareEthMb('192.168.254.167')

    power_board = PowerBoard(engine)
    power_board.read_states()
    print(power_board.dut_1_power.state)
    print(power_board.dut_2_power.state)

    power_board.dut_1_power.state = True
    power_board.dut_2_power.state = False

    power_board.write_states()
    time.sleep(1)

    print(power_board.dut_1_power.state)
    print(power_board.dut_2_power.state)

    power_board.dut_1_power.state = False
    power_board.dut_2_power.state = True

    power_board.write_states()

    print(power_board.dut_1_power.state)
    print(power_board.dut_2_power.state)


"""

from dataclasses import dataclass
from typing import List, Literal

from pymodbus.client import ModbusTcpClient
from pymodbus.exceptions import (
    ModbusException,
    ModbusIOException,
)
from pymodbus.framer.base import FramerType
from retry import retry

from hwc.common import SignalProperties, SignalsEngine


@dataclass
class SignalPropertiesWaveShareEthMb(SignalProperties):
    """Signal Properties specific for a `WaveShare Modbus Poe ETH Relay` module board."""

    relay_no: int
    active_state: Literal[True, False]


class SignalEnginWaveShareEthMb(SignalsEngine):
    """Signals engin implementation for WaveShare Modbus Poe ETH Relay.

    More information: https://www.waveshare.com/wiki/Modbus_POE_ETH_Relay

    """

    _SIGNAL_PROPERTY_TYPE = SignalPropertiesWaveShareEthMb

    def __init__(self, host_ip: str, port=4196, slave=1, number_of_relays: int = 30) -> None:
        super().__init__()
        self._slave = slave
        self._number_of_relays = number_of_relays
        self._modbus = ModbusTcpClient(
            host=host_ip,
            port=port,
            framer=FramerType.RTU,
            reconnect_delay=1,
            retries=3,
        )

    @retry(exceptions=ModbusException, tries=3, delay=1)
    def read_states(self) -> None:
        """Read all signal (relay) states from the board."""
        with self._modbus:
            relays_state = self._modbus.read_coils(
                address=0, count=self._number_of_relays, slave=self._slave
            )
        if isinstance(relays_state, ModbusIOException):
            raise relays_state

        self._update_signals_state(relays_state.bits, self._signal_members)

    def _set_relays_states(self) -> None:
        signals_to_update = [
            i for i in self._signal_members if not i.__state__ == i.__new_state__
        ]
        for signal in signals_to_update:
            try:
                relay_no = signal.get_hw_property_by_type(
                    self._SIGNAL_PROPERTY_TYPE
                ).relay_no
            except ValueError:
                continue
            active_state = signal.get_hw_property_by_type(
                self._SIGNAL_PROPERTY_TYPE
            ).active_state
            new_state = (
                signal.__new_state__ if active_state else not signal.__new_state__
            )
            with self._modbus:
                self._modbus.write_coil(
                    address=relay_no - 1, value=new_state, slave=self._slave
                )

    def _update_signals_state(
        self, current_status: List[bool], signals: List["Signal"]
    ) -> None:
        for signal in signals:
            try:
                relay_no = signal.get_hw_property_by_type(
                    self._SIGNAL_PROPERTY_TYPE
                ).relay_no
            except ValueError:
                continue
            active_state = signal.get_hw_property_by_type(
                self._SIGNAL_PROPERTY_TYPE
            ).active_state

            state = current_status[relay_no - 1]
            state = state if active_state else not state
            signal.__state__ = state

    @retry(exceptions=ModbusException, tries=3, delay=1)
    def write_states(self) -> None:
        """Set all updated signal (relay) states to boards."""
        self._set_relays_states()
        self.read_states()



"""Provide base classes and types for hardware signal representation."""

from abc import ABC, abstractmethod, ABCMeta
from typing import List, TypeVar


class SingletonMeta(ABCMeta):
    """Singleton metaclass for AbstractMeta."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class SignalsEngine(ABC):
    """Abstract class for Signals' Engine."""

    def __init__(self):
        self._signal_members = None

    def associate_signal_members(self, signal_members: List["Signal"]):
        """Assign signal members to engine instance.

        :param signal_members: members of signals class

        """
        self._signal_members = signal_members

    @abstractmethod
    def read_states(self) -> None:
        """Read all signal states using given engine."""

    @abstractmethod
    def write_states(self) -> None:
        """Set all updated signal states using given engine."""


SignalsEnginBase = TypeVar("SignalsEnginBase", bound=SignalsEngine)


class SignalProperties:  # pylint: disable=too-few-public-methods
    """Signal Properties base type."""


SignalPropertiesBase = TypeVar("SignalPropertiesBase", bound=SignalProperties)


class Signal:
    """Base Signal representation."""

    def __init__(self, hardware_properties: List[SignalPropertiesBase]):
        self._hardware_properties = hardware_properties
        self._state = None
        self._new_state = None
        self.__name__ = ""

    def __str__(self) -> str:
        return f"[{self.__class__.__name__}] {self.__name__}.state: {self._state}"

    def get_hw_property_by_type(
        self, hw_property_type: type["SignalProperties"]
    ) -> "SignalProperties":
        """Return hardware device property by given property type.

        :param hw_property_type: type of properties to be returned
        :returns: hardware device signal property

        """
        for hw_property in self._hardware_properties:
            if isinstance(hw_property, hw_property_type):
                return hw_property
        raise ValueError(
            f"There is no `{hw_property_type}` defined in hardware_properties."
        )

    @property
    def __state__(self):
        return self._state

    @__state__.setter
    def __state__(self, value):
        self._state = self._new_state = value

    @property
    def __new_state__(self):
        return self._new_state

    @__new_state__.setter
    def __new_state__(self, value):
        self._new_state = value


class Signals(metaclass=SingletonMeta):
    """Base Signals container."""

    def __init__(self, engine: SignalsEngine) -> None:
        self.__signal_members__: List[Signal] = []
        self._get_signal_members()
        self._engine = engine
        self._engine.associate_signal_members(self.__signal_members__)
        self._next_index = 0

    def _get_signal_members(self) -> None:
        for member_name in dir(self.__class__):
            member = getattr(self, member_name)
            if isinstance(member, Signal):
                member.__name__ = member_name
                self.__signal_members__.append(member)

    def read_states(self) -> None:
        """Read all signal states using given engine."""
        self._engine.read_states()

    def write_states(self) -> None:
        """Set all updated signal states using given engine."""
        self._engine.write_states()

    def __iter__(self):
        self._next_index = 0
        return self

    def __next__(self):
        if self._next_index < len(self.__signal_members__):
            value = self.__signal_members__[self._next_index]
            self._next_index += 1
            return value
        raise StopIteration

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class DISignal(Signal):
    """Digital input signal representation."""

    @property
    def state(self):
        """State of Digital Input signal."""
        return bool(self.__state__)


class DOSignal(Signal):
    """Digital output signal representation."""

    @property
    def state(self) -> bool:
        """State of Digital Output signal."""
        if not self.__state__ == self.__new_state__:
            raise RuntimeError("A new state has not been sent to the device.")
        return bool(self.__state__)

    @state.setter
    def state(self, value: bool) -> None:
        """Set state of Digital Output signal.

        :param value: new value of digital output signal

        """
        self.__new_state__ = value



"""Provide signals support for AVT7510 relays module board.

Example configuration:

.. code-block:: python

    from hwc.common import Signals, DOSignal
    from hwc.avt_5710.signals_avt_5710 import SignalEnginAvt5710, SignalPropertiesAvt5710


    class PowerBoard(Signals):
        dev1_power = DOSignal(
            hardware_properties=[
                SignalPropertiesAvt5710(relay_no=1, active_state=False),
            ]
        )
        dev2_power = DOSignal(
            hardware_properties=[
                SignalPropertiesAvt5710(relay_no=2, active_state=False),
            ]
        )

        dev1_pi = DOSignal(
            hardware_properties=[
                SignalPropertiesAvt5710(relay_no=5, active_state=True),
            ]
        )
        dev2_pi = DOSignal(
            hardware_properties=[
                SignalPropertiesAvt5710(relay_no=6, active_state=True),
            ]
        )


    avt_engine = SignalEnginAvt5710("COM5")

    power_board = PowerBoard(avt_engine)
    power_board.read_states()
    print(power_board.dev1_power.state)

    power_board.dev1_power.state = False

    power_board.write_states()

    print(power_board.dev1_power.state)


"""
import time
from dataclasses import dataclass
from typing import List, Literal

import serial
from retry import retry

from hwc.common import SignalProperties, SignalsEngine

# pylint notice similarity between AVT and waveshare implementations
# but there's no reason to extract it to base class or make it a template method
# pylint:disable=R0801

@dataclass
class SignalPropertiesAvt5710(SignalProperties):
    """Signal Properties specific for a AVT5710 relays module board."""

    relay_no: int
    active_state: Literal[True, False]


class SignalEnginAvt5710(SignalsEngine):
    """Signals engin implementation for AVT5710.

    The board is controlled via a serial COM port. Transmission parameters: 19200 8N1.
    Commands are sent as ASCII characters and always begin with the start character `Esc`(0x1B)
    and finish with the end character `Enter` (0x0D).

    More information: https://serwis.avt.pl/manuals/AVT5710.pdf

    """

    _SIGNAL_PROPERTY_TYPE = SignalPropertiesAvt5710
    _START_CHR = b"\x1B"  # Esc
    _STOP_CHR = b"\x0D"  # Enter
    _SET_CMD = b"S"
    _RST_CMD = b"C"
    _COM_BAUD_RATE = 19200
    _READ_K_STATE = b"L"
    _READ_ALL_STATE = b"R"

    def __init__(self, com_port: str) -> None:
        super().__init__()
        self._avt_serial = self._init_serial_port(com_port)

    @retry(serial.serialutil.SerialException, tries=5, delay=1)
    def _init_serial_port(self, com_port: str):
        return serial.Serial(com_port, self._COM_BAUD_RATE, timeout=1)

    @retry((RuntimeError, serial.serialutil.SerialException), tries=10, delay=1)
    def _read_states(self) -> int:
        with self._avt_serial as com_port:
            cmd = self._START_CHR + self._READ_ALL_STATE + self._STOP_CHR
            com_port.write(cmd)
            state = com_port.read(4)

        if (
            bytes([state[0]]) == self._READ_ALL_STATE
            and bytes([state[3]]) == self._STOP_CHR
        ):
            state_as_bytes = bytes.fromhex(chr(state[1]) + chr(state[2]))
            return int.from_bytes(state_as_bytes)

        raise RuntimeError(
            f"Can not read relays state of AVT5710 module. Received response: {state}."
        )

    @retry((RuntimeError, serial.serialutil.SerialException), tries=10, delay=1)
    def _set_relay_states(self, relay_no: int, state: bool) -> None:
        relay_no_ascii = ord(str(relay_no)).to_bytes()
        ctrl_cmd = self._SET_CMD if state else self._RST_CMD
        with self._avt_serial as com_port:
            cmd = self._START_CHR + ctrl_cmd + relay_no_ascii + self._STOP_CHR
            try:
                com_port.write(cmd)
            except TimeoutError as e:
                raise RuntimeError(
                    f"Can not write relay: {relay_no} state to AVT5710 module."
                ) from e
            time.sleep(0.1)

    def read_states(self) -> None:
        """Read all signal (relay) states from the AVT5710 board"""
        relays_state = self._read_states()
        self._update_signals_state(relays_state, self._signal_members)

    def _set_relays_states(self) -> None:
        signals_to_update = [
            i for i in self._signal_members if not i.__state__ == i.__new_state__
        ]
        for signal in signals_to_update:
            try:
                relay_no = signal.get_hw_property_by_type(
                    self._SIGNAL_PROPERTY_TYPE
                ).relay_no
            except ValueError:
                continue
            active_state = signal.get_hw_property_by_type(
                self._SIGNAL_PROPERTY_TYPE
            ).active_state
            new_state = (
                signal.__new_state__ if active_state else not signal.__new_state__
            )
            self._set_relay_states(relay_no, new_state)

    def _update_signals_state(
        self, current_status: int, signals: List["Signal"]
    ) -> None:
        for signal in signals:
            try:
                relay_no = signal.get_hw_property_by_type(
                    self._SIGNAL_PROPERTY_TYPE
                ).relay_no
            except ValueError:
                continue
            active_state = signal.get_hw_property_by_type(
                self._SIGNAL_PROPERTY_TYPE
            ).active_state

            state = bool(current_status & (2 ** (relay_no - 1)))
            state = state if active_state else not state
            signal.__state__ = state

    def write_states(self) -> None:
        """Set all updated signal (relay) states to AVT5710 boards."""
        self._set_relays_states()
        self.read_states()
